"""

Data source: http://www.railwaycodes.org.uk

ELRs, mileages, operators, grid references
Bilingual station names
Sponsored stations
Stations not served by their Station Facility Operator (SFO)
International stations
Station trivia

Railway station data (http://www.railwaycodes.org.uk/stations/station0.shtm)

"""

import copy
import itertools
import os
import re
import string
import urllib.parse

import bs4
import numpy as np
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.store import load_pickle

from pyrcs.utils import cd_dat, get_catalogue, get_last_updated_date, parse_location_note, parse_table
from pyrcs.utils import save_pickle


class Stations:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Railway station data'
        self.URL = self.HomeURL + '/stations/station0.shtm'
        self.Catalogue = get_catalogue(self.URL)

        source = requests.get(self.URL)
        soup = bs4.BeautifulSoup(source.text, 'lxml').find_all('a', href=True)
        self.SubCatalogue = dict((x.text, urllib.parse.urljoin(os.path.dirname(self.URL) + '/', x['href']))
                                 for x in [x for x in soup if x.text in string.ascii_uppercase][2:])

        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("other-assets", "stations")
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "dat\\other-assets\\stations" and sub-directories
    def cd_stn(self, *sub_dir):
        path = self.DataDir
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\other-assets\\stations\\dat"
    def cdd_stn(self, *sub_dir):
        path = self.cd_stn("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Parse 'Operator' column
    @staticmethod
    def parse_current_operator(x):
        contents = re.split(r'\\r| \[\'|\\\\r| {2}\'\]|\', \'|\\n',
                            x.lstrip(' [\'').rstrip('  \']').lstrip('\n').strip())
        contents = [x for x in contents if x != '']
        operators = []
        for y in contents:
            # Operators names
            operator_name = re.search(r'.*(?= \(from \d+ \w+ \d+(.*)?\))', y)
            operator_name = operator_name.group() if operator_name is not None else ''
            # Start dates
            start_date = re.search(r'(?<= \(from )\d+ \w+ \d+( to \d+ \w+ \d+(.*))?(?=\))', y)
            start_date = start_date.group() if start_date is not None else ''
            # Form a tuple
            operators.append((operator_name, start_date))
        return operators

    # Collect railway station data for a given 'keyword'
    def collect_station_locations(self, initial, update=False, verbose=False):
        """
        :param initial: [str] station data (including the station name, ELR, mileage, status, owner, operator,
                            degrees of longitude and latitude, and grid reference) for stations whose name start with
        :param update: [bool]
        :param verbose: [bool]
        :return [dict] {keyword: [DataFrame] railway station data,
                        'Last_updated_date': [str] date of when the data was last updated}
        """
        path_to_pickle = self.cd_stn("a-z", initial.lower() + ".pickle")

        if os.path.isfile(path_to_pickle) and not update:
            station_location_codes = load_pickle(path_to_pickle)

        else:
            url = self.URL.replace('station0', 'station{}'.format(initial.lower()))

            if initial.upper() not in list(self.SubCatalogue.keys()):
                print("No data is available for signal box codes beginning with \"{}\".".format(initial.upper()))
                station_location_codes = {initial.upper(): pd.DataFrame(), 'Last_updated_date': ''}

            else:
                try:
                    source = requests.get(url)  # Request to get connected to the url
                    records, header = parse_table(source, parser='lxml')
                    # Create a DataFrame of the requested table
                    dat = [[x.replace('=', 'See').strip('\xa0') for x in i] for i in records]
                    col = [re.sub(r'\n?\r+\n?', ' ', h) for h in header]
                    station_locations_table = pd.DataFrame(dat, columns=col)

                    def parse_degrees(x):
                        if x == '':
                            y = np.nan
                        else:
                            y = float(x.replace('c.', '') if x.startswith('c.') else x)
                        return y

                    station_locations_table[['Degrees Longitude', 'Degrees Latitude']] = \
                        station_locations_table[['Degrees Longitude', 'Degrees Latitude']].applymap(parse_degrees)
                    station_locations_table['Grid Reference'] = station_locations_table['Grid Reference'].map(
                        lambda x: x.replace('c.', '') if x.startswith('c.') else x)

                    station_locations_table[['Station', 'Station_Note']] = \
                        station_locations_table.Station.map(parse_location_note).apply(pd.Series)

                    # Operator
                    temp = list(station_locations_table.Operator.map(self.parse_current_operator))
                    length = len(max(temp, key=len))
                    col_names_current = ['Operator', 'Date']
                    prev_no = list(
                        itertools.chain.from_iterable(itertools.repeat(x, 2) for x in list(range(1, length))))
                    col_names = zip(col_names_current * (length - 1), prev_no)
                    col_names = col_names_current + ['_'.join(['Prev', x, str(d)]) for x, d in col_names]

                    for i in range(len(temp)):
                        if len(temp[i]) < length:
                            temp[i] += [(None, None)] * (length - len(temp[i]))

                    temp = pd.DataFrame(temp)
                    operators = [pd.DataFrame(temp)[col].apply(pd.Series) for col in temp.columns]
                    operators = pd.concat(operators, axis=1, sort=False)
                    operators.columns = col_names

                    station_locations_table.drop('Operator', axis=1, inplace=True)
                    station_locations_table = station_locations_table.join(operators)

                except Exception as e:
                    print("Failed to collect station location codes beginning with \"{}\". {}".format(
                        initial.upper(), e))
                    station_locations_table = pd.DataFrame()

                try:
                    last_updated_date = get_last_updated_date(url)
                except Exception as e:
                    print("Failed to find the last updated date of the station location codes beginning with "
                          "\"{}\" {}".format(initial.upper(), e))
                    last_updated_date = ''

                station_location_codes = {initial.upper(): station_locations_table,
                                          'Last_updated_date': last_updated_date}

            save_pickle(station_location_codes, path_to_pickle, verbose)

        return station_location_codes

    # Fetch all of the collected railway station data
    def fetch_station_locations(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        """
        :param update: [bool]
        :param pickle_it: [bool]
        :param data_dir: [str; None]
        :param verbose: [bool]
        :return [dict] {initial: [DataFrame] station data, including the station name, ELR, mileage, status, owner,
                                    operator, degrees of longitude and latitude, and grid reference,
                        'Latest_update_date': [str] date of when the data was last updated}
        """
        data_sets = [self.collect_station_locations(x, update, verbose=False if data_dir or not verbose else True)
                     for x in string.ascii_lowercase]

        station_location_codes_tables = (item[x] for item, x in zip(data_sets, string.ascii_uppercase))
        station_location_codes = pd.concat(station_location_codes_tables, axis=0, ignore_index=True, sort=False)

        last_updated_dates = (d['Last_updated_date'] for d in data_sets)
        latest_update_date = max(d for d in last_updated_dates if d != '')

        station_locations_data = {'Stations': station_location_codes, 'Latest_update_date': latest_update_date}

        if pickle_it and data_dir:
            self.CurrentDataDir = regulate_input_data_dir(data_dir)
            path_to_pickle = os.path.join(self.CurrentDataDir, "station_locations.pickle")
            save_pickle(station_locations_data, path_to_pickle, verbose=True)

        return station_locations_data



""" 
Data source: http://www.railwaycodes.org.uk

CRS, NLC, TIPLOC and STANOX Codes (http://www.railwaycodes.org.uk/crs/CRS0.shtm)

This links to a four-way listing of railway codes:

    - Computer reservation system (CRS) codes
        * [replaced by national reservation system (NRS) codes from late 2004, but the codes are the same]
    - National location codes (NLC)
    - Timing point locations (TIPLOC)
    - Station number names (STANME)
    - Station numbers (STANOX)
"""

import copy
import os
import re
import string
import urllib.parse

import bs4
import more_itertools
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.ops import confirmed
from pyhelpers.store import load_json, load_pickle

from pyrcs.utils import cd_dat, save_json, save_pickle
from pyrcs.utils import get_catalogue, get_last_updated_date
from pyrcs.utils import parse_date, parse_location_note, parse_table, parse_tr


class LocationIdentifiers:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'CRS, NLC, TIPLOC and STANOX codes'
        self.URL = self.HomeURL + '/crs/CRS0.shtm'
        self.Catalogue = get_catalogue(self.URL)
        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("line-data", "crs-nlc-tiploc-stanox")
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "dat\\line-data\\crs-nlc-tiploc-stanox\\" and sub-directories
    def cd_lc(self, *sub_dir):
        path = self.DataDir
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\line-data\\crs-nlc-tiploc-stanox\\dat" and sub-directories
    def cdd_lc(self, *sub_dir):
        path = self.cd_lc("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Location name modifications
    @staticmethod
    def amendment_to_location_names():
        location_name_amendment_dict = {
            'Location': {re.compile(r' And | \+ '): ' & ',
                         re.compile(r'-By-'): '-by-',
                         re.compile(r'-In-'): '-in-',
                         re.compile(r'-En-Le-'): '-en-le-',
                         re.compile(r'-La-'): '-la-',
                         re.compile(r'-Le-'): '-le-',
                         re.compile(r'-On-'): '-on-',
                         re.compile(r'-The-'): '-the-',
                         re.compile(r' Of '): ' of ',
                         re.compile(r'-Super-'): '-super-',
                         re.compile(r'-Upon-'): '-upon-',
                         re.compile(r'-Under-'): '-under-',
                         re.compile(r'-Y-'): '-y-'}}
        return location_name_amendment_dict

    # Parse addition note page
    @staticmethod
    def parse_additional_note_page(url, parser='lxml'):
        source = requests.get(url)
        web_page_text = bs4.BeautifulSoup(source.text, parser).find_all(['p', 'pre'])
        parsed_text = [x.text for x in web_page_text if isinstance(x.next_element, str)]
        parsed_texts = []
        for x in parsed_text:
            if '\n' in x:
                text = re.sub('\t+', ',', x).replace('\t', ' ').replace('\xa0', '').split('\n')
            else:
                text = x.replace('\t', ' ').replace('\xa0', '')
            if isinstance(text, list):
                text = [t.split(',') for t in text if t != '']
                parsed_texts.append(pd.DataFrame(text, columns=['Location', 'CRS', 'CRS_alt1', 'CRS_alt2']).fillna(''))
            else:
                to_remove = ['click the link', 'click your browser', 'Thank you', 'shown below']
                if text != '' and not any(t in text for t in to_remove):
                    parsed_texts.append(text)
        return parsed_texts

    # Collect note about CRS
    def collect_additional_crs_note(self, confirmation_required=True, verbose=False):
        if confirmed("To collect additional CRS note?", confirmation_required=confirmation_required):

            try:
                note_url = self.HomeURL + '/crs/CRS2.shtm'
                additional_note = self.parse_additional_note_page(note_url)
                additional_crs_note, notes = {}, []
                for x in additional_note:
                    if isinstance(x, str):
                        if 'Last update' in x:
                            additional_crs_note.update({'Last_updated_date': parse_date(x, as_date_type=False)})
                        else:
                            notes.append(x)
                    else:
                        additional_crs_note.update({'Alternative_CRS': x})
                additional_crs_note.update({'Note': notes})
                save_pickle(additional_crs_note, self.cd_lc("additional-crs-note.pickle"), verbose)

            except Exception as e:
                print("Failed to collect/update additional note for CRS. {}.".format(e))
                additional_crs_note = None

            return additional_crs_note

    # Fetch note about CRS
    def fetch_additional_crs_note(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        pickle_filename = "additional-crs-note.pickle"
        path_to_pickle = self.cd_lc(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            additional_note = load_pickle(path_to_pickle)
        else:
            additional_note = self.collect_additional_crs_note(confirmation_required=False,
                                                               verbose=False if data_dir or not verbose else True)
            if additional_note:  # additional_note is not None
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
                    save_pickle(additional_note, path_to_pickle, verbose=True)
            else:
                print("No data of additional note for CRS has been collected.")
        return additional_note

    # Collect data for other systems
    def collect_other_systems_codes(self, confirmation_required=True, verbose=False):
        if confirmed("To collect other systems codes?", confirmation_required=confirmation_required):

            try:
                source = requests.get(self.Catalogue['Other systems'])
                web_page_text = bs4.BeautifulSoup(source.text, 'lxml')
                # Get system name
                system_names = [k.text for k in web_page_text.find_all('h3')]
                system_names = [n.replace('Tramlnk', 'Tramlink') if 'Tramlnk' in n else n for n in system_names]
                # Get column names for the other systems table
                headers = list(more_itertools.unique_everseen([h.text for h in web_page_text.find_all('th')]))
                # Parse table data for each system
                tbl_data = web_page_text.find_all('table')
                tables = [pd.DataFrame(parse_tr(headers, table.find_all('tr')), columns=headers) for table in tbl_data]
                codes = [tables[i] for i in range(len(tables)) if i % 2 != 0]
                # Make a dict
                other_systems_codes = dict(zip(system_names, codes))
                save_pickle(other_systems_codes, self.cd_lc("other-systems-codes.pickle"), verbose)

            except Exception as e:
                print("Failed to collect location codes for other systems. {}.".format(e))
                other_systems_codes = None

            return other_systems_codes

    # Fetch the data for other systems
    def fetch_other_systems_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        pickle_filename = "other-systems-codes.pickle"
        path_to_pickle = self.cd_lc(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            other_systems_codes = load_pickle(path_to_pickle)

        else:
            other_systems_codes = self.collect_other_systems_codes(confirmation_required=False,
                                                                   verbose=False if data_dir or not verbose else True)
            if other_systems_codes:  # other_systems_codes is not None
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
                    save_pickle(other_systems_codes, path_to_pickle, verbose=True)
            else:
                print("No data of other systems codes has been collected.")

        return other_systems_codes

    # Collect data of locations including CRS, NLC, TIPLOC, STANME and STANOX codes
    def collect_location_codes_by_initial(self, initial, update=False, verbose=False):
        """
        :param initial: [str] initial letter of station/junction name or certain word for specifying URL
        :param update: [bool]
        :param verbose: [bool]
        :return [tuple] ([DataFrame] CRS, NLC, TIPLOC and STANOX data of (almost) all stations/junctions,
                         [str]} date of when the data was last updated)
        """
        assert initial in string.ascii_letters
        path_to_pickle = self.cd_lc("a-z", initial.lower() + ".pickle")

        if os.path.isfile(path_to_pickle) and not update:
            location_codes_data = load_pickle(path_to_pickle)

        else:
            url = self.Catalogue[initial.upper()]

            try:
                last_updated_date = get_last_updated_date(url)
            except Exception as e:
                print("Failed to find the last update date for codes starting with \"{}.\" {}".format(
                    initial.upper(), e))
                last_updated_date = ''

            try:
                source = requests.get(url)  # Request to get connected to the URL
                tbl_lst, header = parse_table(source, parser='lxml')

                # Get a raw DataFrame
                reps = {'\b-\b': '', '\xa0\xa0': ' ', '&half;': ' and 1/2'}
                pattern = re.compile("|".join(reps.keys()))
                tbl_lst = [[pattern.sub(lambda x: reps[x.group(0)], item) for item in record] for record in tbl_lst]
                location_codes = pd.DataFrame(tbl_lst, columns=header)
                location_codes.replace({'\xa0': ''}, regex=True, inplace=True)

                # Collect additional information as note
                location_codes[['Location', 'Location_Note']] = location_codes.Location.map(
                    parse_location_note).apply(pd.Series)

                # CRS, NLC, TIPLOC, STANME
                drop_pattern = re.compile(r'[Ff]ormerly|[Ss]ee[ also]|Also .[\w ,]+')
                idx = [location_codes[location_codes.CRS == x].index[0]
                       for x in location_codes.CRS if re.match(drop_pattern, x)]
                location_codes.drop(labels=idx, axis=0, inplace=True)

                # Collect others note
                def collect_others_note(other_note_x):
                    n = re.search(r'(?<=[\[(\'])[\w,? ]+(?=[)\]\'])', other_note_x)
                    note = n.group() if n is not None else ''
                    return note

                # Strip others note
                def strip_others_note(other_note_x):
                    d = re.search(r'[\w ,]+(?= [\[(\'])', other_note_x)
                    dat = d.group() if d is not None else other_note_x
                    return dat

                other_codes_col = location_codes.columns[1:-1]
                other_notes_col = [x + '_Note' for x in other_codes_col]
                location_codes[other_notes_col] = location_codes[other_codes_col].applymap(collect_others_note)
                location_codes[other_codes_col] = location_codes[other_codes_col].applymap(strip_others_note)

                # Parse STANOX note
                def parse_stanox_note(x):
                    if x == '-':
                        dat, note = '', ''
                    else:
                        d = re.search(r'[\w *,]+(?= [\[(\'])', x)
                        dat = d.group() if d is not None else x
                        note = 'Pseudo STANOX' if '*' in dat else ''
                        n = re.search(r'(?<=[\[(\'])[\w, ]+.(?=[)\]\'])', x)
                        if n is not None:
                            note = '; '.join(x for x in [note, n.group()] if x != '')
                        if '(' not in note and note.endswith(')'):
                            note = note.rstrip(')')
                        dat = dat.rstrip('*') if '*' in dat else dat
                    return dat, note

                if not location_codes.empty:
                    location_codes[['STANOX', 'STANOX_Note']] = location_codes.STANOX.map(
                        parse_stanox_note).apply(pd.Series)
                else:  # It is likely that no data is available on the web page for the given 'key_word'
                    location_codes['STANOX_Note'] = location_codes.STANOX

                if any('see note' in crs_note for crs_note in location_codes.CRS_Note):
                    loc_idx = [i for i, crs_note in enumerate(location_codes.CRS_Note) if 'see note' in crs_note]
                    web_page_text = bs4.BeautifulSoup(source.text, 'lxml')
                    note_urls = [urllib.parse.urljoin(self.Catalogue[initial.upper()], x['href'])
                                 for x in web_page_text.find_all('a', href=True, text='note')]
                    additional_notes = [self.parse_additional_note_page(note_url) for note_url in note_urls]
                    additional_note = dict(zip(location_codes.CRS.iloc[loc_idx], additional_notes))
                else:
                    additional_note = None

                location_codes = location_codes.replace(self.amendment_to_location_names(), regex=True)

                location_codes.STANOX = location_codes.STANOX.replace({'-': ''})

                location_codes.index = range(len(location_codes))  # Rearrange index

            except Exception as e:
                print("Failed to collect the codes of locations starting with \"{}\". {}.".format(initial.upper(), e))
                location_codes, additional_note = pd.DataFrame(), None

            location_codes_data = dict(zip([initial.upper(), 'Additional_note', 'Last_updated_date'],
                                           [location_codes, additional_note, last_updated_date]))

            save_pickle(location_codes_data, path_to_pickle, verbose)

        return location_codes_data

    # Fetch all Location data including CRS, NLC, TIPLOC, STANME and STANOX codes either locally or from online
    def fetch_location_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        # Get every data table
        data = [self.collect_location_codes_by_initial(x, update, verbose=False if data_dir or not verbose else True)
                for x in string.ascii_lowercase]

        # Select DataFrames only
        location_codes_data = (item[x] for item, x in zip(data, string.ascii_uppercase))
        location_codes_data_table = pd.concat(location_codes_data, axis=0, ignore_index=True, sort=False)

        # Likely errors (spotted occasionally)
        idx = location_codes_data_table[location_codes_data_table.Location == 'Selby Melmerby Estates'].index
        values = location_codes_data_table.loc[idx, 'STANME':'STANOX'].values
        location_codes_data_table.loc[idx, 'STANME':'STANOX'] = ['', '']
        idx = location_codes_data_table[location_codes_data_table.Location == 'Selby Potter Group'].index
        location_codes_data_table.loc[idx, 'STANME':'STANOX'] = values

        # Get the latest updated date
        last_updated_dates = (item['Last_updated_date'] for item, _ in zip(data, string.ascii_uppercase))
        latest_update_date = max(d for d in last_updated_dates if d is not None)

        # Get additional note
        additional_note = self.fetch_additional_crs_note()

        # Get other systems codes
        other_systems_codes = self.fetch_other_systems_codes()

        # Create a dict to include all information
        location_codes = {'Location_codes': location_codes_data_table,
                          'Latest_update_date': latest_update_date,
                          'Additional_note': additional_note,
                          'Other_systems': other_systems_codes}

        if pickle_it and data_dir:
            self.CurrentDataDir = regulate_input_data_dir(data_dir)
            path_to_pickle = os.path.join(self.CurrentDataDir, "crs-nlc-tiploc-stanox-codes.pickle")
            save_pickle(location_codes, path_to_pickle, verbose=True)

        return location_codes

    # Make a dict/dataframe for location code data for the given keyword
    def make_location_codes_dictionary(self, keys, initials=None, main_key=None, drop_duplicates=False, as_dict=False,
                                       pickle_it=False, data_dir=None, update=False):
        """
        :param keys: [list] e.g. ['CRS', 'NLC', 'TIPLOC', 'STANOX', 'STANME']
        :param initials: [str] one of string.ascii_letters
        :param main_key: [str; None]
        :param drop_duplicates: [bool] If drop_duplicates is False, loc_dict will take the last item to be the value
        :param as_dict: [bool]
        :param pickle_it: [bool]
        :param data_dir: [str; None]
        :param update: [bool]
        :return:
        """
        if isinstance(keys, str):
            assert keys in ('CRS', 'NLC', 'TIPLOC', 'STANOX', 'STANME')
            keys = [keys]
        elif isinstance(keys, list):
            assert all(x in ('CRS', 'NLC', 'TIPLOC', 'STANOX', 'STANME') for x in keys)
        if isinstance(initials, str):
            assert initials in string.ascii_letters
            initials = [initials]
        elif isinstance(initials, list):
            assert all(x in string.ascii_letters for x in initials)
        else:
            assert initials is None
        assert isinstance(main_key, str) or main_key is None

        dat_dir = regulate_input_data_dir(data_dir) if data_dir else self.DataDir
        path_to_file = os.path.join(dat_dir, "-".join(keys) +
                                    ("" if initials is None else "-" + "".join(initials)) +
                                    (".json" if as_dict and len(keys) == 1 else ".pickle"))

        if os.path.isfile(path_to_file) and not update:
            if as_dict:
                location_codes_dictionary = load_json(path_to_file)
            else:
                location_codes_dictionary = load_pickle(path_to_file)

        else:
            if initials is None:
                location_codes = self.fetch_location_codes()['Location_codes']
            else:
                temp = [self.collect_location_codes_by_initial(initial)[initial.upper()] for initial in initials]
                location_codes = pd.concat(temp, axis=0, ignore_index=True, sort=False)
            assert isinstance(location_codes, pd.DataFrame)

            # Deep cleansing location_code
            try:
                key_location_codes = location_codes[['Location'] + keys]
                key_location_codes = key_location_codes.query(' | '.join(['{} != \'\''.format(k) for k in keys]))

                #
                if drop_duplicates:
                    location_codes_subset = key_location_codes.drop_duplicates(subset=keys, keep='first')
                    location_codes_duplicated = None

                else:  # drop_duplicates is False or None
                    location_codes_subset = key_location_codes.drop_duplicates(subset=keys, keep=False)
                    #
                    dupl_temp_1 = key_location_codes[key_location_codes.duplicated(['Location'] + keys, keep=False)]
                    dupl_temp_2 = key_location_codes[key_location_codes.duplicated(keys, keep=False)]
                    duplicated_1 = dupl_temp_2[dupl_temp_1.eq(dupl_temp_2)].dropna().drop_duplicates()
                    duplicated_2 = dupl_temp_2[~dupl_temp_1.eq(dupl_temp_2)].dropna()
                    duplicated = pd.concat([duplicated_1, duplicated_2], axis=0, sort=False)
                    location_codes_duplicated = duplicated.groupby(keys).agg(tuple)
                    location_codes_duplicated.Location = location_codes_duplicated.Location.map(
                        lambda x: x[0] if len(set(x)) == 1 else x)

                location_codes_subset.set_index(keys, inplace=True)
                location_codes_ref = pd.concat([location_codes_subset, location_codes_duplicated], axis=0, sort=False)

                if as_dict:
                    location_codes_ref_dict = location_codes_ref.to_dict()
                    if main_key is not None:
                        location_codes_ref_dict[main_key] = location_codes_ref_dict.pop('Location')
                        location_codes_dictionary = location_codes_ref_dict
                    else:
                        location_codes_dictionary = location_codes_ref_dict['Location']
                else:

                    location_codes_dictionary = location_codes_ref

                if pickle_it:
                    if path_to_file.endswith(".json"):
                        save_json(location_codes_dictionary, path_to_file)
                    else:
                        save_pickle(location_codes_dictionary, path_to_file)

            except Exception as e:
                print("Failed to get multiple location code indexed reference. {}.".format(e))
                location_codes_dictionary = None

        return location_codes_dictionary



import os

import setuptools

ver_no = '0.2.6'


def find_all_pkg_dat_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


dat_files = find_all_pkg_dat_files('pyrcs\\dat')


with open("README.md", 'r', encoding='utf-8') as readme:
    long_description = readme.read()

with open('requirements.txt') as f:
    requirements = f.readlines()
requirements_ = [r.strip() for r in requirements]

setuptools.setup(

    name='pyrcs',
    version=ver_no,

    author='Qian Fu',
    author_email='qian.fu@outlook.com',

    description="A data scraping tool for collection and storage of the railway codes used in the UK rail industry",
    long_description=long_description,
    long_description_content_type="text/markdown",

    url='https://github.com/mikeqfu/pyrcs',

    install_requires=requirements_,

    packages=setuptools.find_packages(exclude=["*.tests", "tests.*", "tests"]),

    package_data={"pyrcs": dat_files + ["line_data_cls/*", "other_assets_cls/*"]},
    include_package_data=True,

    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux'
    ],
)



""" Update package data """

import time

from pyhelpers.ops import confirmed

from pyrcs.line_data import LineData
from pyrcs.other_assets import OtherAssets


def update_pkg_metadata(verbose=False):

    if confirmed("To update resources?"):

        line_dat = LineData()

        # ELR and mileages
        _ = line_dat.ELRMileages.fetch_elr(update=True, verbose=verbose)

        time.sleep(10)

        # Electrification
        _ = line_dat.Electrification.fetch_electrification_codes(update=True, verbose=verbose)

        time.sleep(10)

        # Location
        _ = line_dat.LocationIdentifiers.fetch_location_codes(update=True, verbose=verbose)
        _ = line_dat.LocationIdentifiers.fetch_additional_crs_note(update=True, verbose=verbose)
        _ = line_dat.LocationIdentifiers.fetch_other_systems_codes(update=True, verbose=verbose)

        time.sleep(10)

        # Line of routes
        _ = line_dat.LOR.fetch_lor_codes(update=True, verbose=verbose)
        _ = line_dat.LOR.fetch_elr_lor_converter(update=True, verbose=verbose)

        time.sleep(10)

        # Line names
        _ = line_dat.LineNames.fetch_line_names(update=True, verbose=verbose)

        """
        # Track diagrams
        """

        time.sleep(10)

        other_assets = OtherAssets()

        # Signal boxes
        _ = other_assets.SignalBoxes.fetch_signal_box_prefix_codes(update=True, verbose=verbose)
        _ = other_assets.SignalBoxes.fetch_non_national_rail_codes(update=True, verbose=verbose)

        time.sleep(10)

        # Tunnels
        _ = other_assets.Tunnels.fetch_railway_tunnel_lengths(update=True, verbose=verbose)

        time.sleep(10)

        # Viaducts
        _ = other_assets.Viaducts.fetch_railway_viaducts(update=True, verbose=verbose)

        time.sleep(10)

        # Stations
        _ = other_assets.Stations.fetch_station_locations(update=True, verbose=verbose)

        time.sleep(10)

        # Depots
        _ = other_assets.Depots.fetch_depot_codes(update=True, verbose=verbose)

        if verbose:
            print("\nUpdate finished.")


# update_pkg_metadata(verbose=True)
